@@ -2,7 +2,7 @@
  Authors 
  Martin Schlather, schlather@math.uni-mannheim.de
 
- Copyright (C) 2015 -- 2016 Martin Schlather, Reinhard Furrer
+ Copyright (C) 2015 -- 2017 Martin Schlather
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -20,8 +20,13 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
 #include <R_ext/Lapack.h>
-#include "General_utils.h"
-#include "kleinkram.h"
+//#include "def.h" // never change this line
+#include "General_utils.h" //#include <General_utils.h>
+#include "zzz_RandomFieldsUtils.h"
+
+
+#define SCALAR(A,B,C) scalarX(A,B,C, SCALAR_AVX)
+#define LINEAR(A,B,C,D) linearX(A,B,C,D,6)
 
 
 void strcopyN(char *dest, const char *src, int n) {
@@ -32,105 +37,106 @@ void strcopyN(char *dest, const char *src, int n) {
   dest[n] = '\0';
 }
 
-
-double scalar(double *A, double *B, int N) {
-  double ANS;
-  SCALAR_PROD(A, B, N, ANS);
-  return ANS;
-}
-
-void AtA(double *a, int nrow, int ncol, double *A) {
-  // A =  a^T %*% a
-  int i, k, j,
-    dimSq = ncol * ncol;
-  
-  for (k=i=0; i<dimSq; i+=ncol) {
-    for (j=0; j<dimSq; j+=ncol, k++) {
-      A[k] = scalar(a + i, a + j, nrow);
+void AtA(double *a, int nrow, int ncol, double *C) {
+  // C =  A^T %*% A
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) if (MULTIMINSIZE(ncol)) schedule(dynamic, 20)
+#endif  
+  for (int i=0; i<ncol; i++) {
+    double 
+      *A = a + i * nrow,
+      *B = A;
+    for (int j=i; j<ncol; j++, B+=nrow) {
+      C[i * ncol + j] = C[i + ncol * j] = SCALAR(A, B, nrow);
     }
   }
 }
  
 
-void xA(double *x, double*A, int nrow, int ncol, double *y) {
-  int i;
+void xA_noomp(double *x, double*A, int nrow, int ncol, double *y) {
   if (A == NULL) {
     if (nrow != ncol || nrow <= 0) BUG;
     MEMCOPY(y, x, sizeof(double) * nrow);
   } else {
-    for (i=0; i<ncol; i++, A+=nrow) y[i] = scalar(x, A, nrow);
+    for (int i=0; i<ncol; i++) {
+      y[i] = SCALAR(x, A + i * nrow, nrow);
+    }
   }
 }
 
-
-#define TWOSCALAR_PROD(A1, A2, B, N, ANS1, ANS2) {	\
-    int  k_ =0,						\
-      end_ = N - 5;					\
-    ANS1 = ANS2 = 0.0;						\
-    for (; k_<end_; k_+=5) {					\
-      ANS1 += A1[k_] * B[k_]					\
-	+ A1[k_ + 1] * B[k_ + 1]				\
-	+ A1[k_ + 2] * B[k_ + 2]				\
-	+ A1[k_ + 3] * B[k_ + 3]				\
-	+ A1[k_ + 4] * B[k_ + 4];				\
-      ANS2 += A2[k_] * B[k_]					\
-	+ A2[k_ + 1] * B[k_ + 1]				\
-	+ A2[k_ + 2] * B[k_ + 2]				\
-	+ A2[k_ + 3] * B[k_ + 3]				\
-	+ A2[k_ + 4] * B[k_ + 4];				\
-    }								\
-    for (; k_<N; k_++) {					\
-      ANS1 += A1[k_] * B[k_];					\
-      ANS2 += A2[k_] * B[k_];					\
-    }								\
-  }
-
-
+void xA(double *x, double*A, int nrow, int ncol, double *y) {
+  if (A == NULL) {
+    if (nrow != ncol || nrow <= 0) BUG;
+    MEMCOPY(y, x, sizeof(double) * nrow);
+  } else {
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) if (MULTIMINSIZE(ncol) && MULTIMINSIZE(nrow))
+#endif  
+   for (int i=0; i<ncol; i++) y[i] = SCALAR(x, A + i * nrow, nrow);
+  } 
+} 
+  
 void xA(double *x1, double *x2,  double*A, int nrow, int ncol, double *y1,
 	double *y2) {
-  int i;
   if (A == NULL) {
     if (nrow != ncol || nrow <= 0) BUG;
     MEMCOPY(y1, x1, sizeof(double) * nrow);
     MEMCOPY(y2, x2, sizeof(double) * nrow);
   } else {
-    for (i=0; i<ncol; i++, A+=nrow) {
-      double d1, d2;
-      TWOSCALAR_PROD(x1, x2, A, nrow, d1, d2);
-      y1[i] = d1;
-      y2[i] = d2;
+    double *a = A;
+    for (int i=0; i<ncol; i++, a += nrow) {
+      y1[i] = SCALAR(x1, a, nrow);
+      y2[i] = SCALAR(x2, a, nrow);
     }
   }	
 }
 
+void xAx(double *x, double*A, int nrow,  double *y) {
+  double sum = 0.0;
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) reduction(+:sum) if (MULTIMINSIZE(nrow) && MULTIMINSIZE(nrow))
+#endif  
+  for (int i=0; i<nrow; i++) sum += x[i] * SCALAR(x, A + i * nrow, nrow);
+  *y = sum;
+}
 
 void Ax(double *A, double*x, int nrow, int ncol, double *y) {
-  int i,j,k;
   if (A == NULL) {
     if (nrow != ncol || nrow <= 0) BUG;
     MEMCOPY(y, x, sizeof(double) * nrow);
   } else {
-    for (i=0; i<nrow; i++) y[i]=0.0;
-    for (k=i=0; i<ncol; i++) { 
-      for (j=0; j<nrow; j++) {
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) if (MULTIMINSIZE(ncol) && MULTIMINSIZE(nrow))
+    for (int j=0; j<nrow; j++) {
+      double dummy = 0.0;
+      int k = j;
+      for (int i=0; i<ncol; i++, k+=nrow) { 
+	dummy += A[k] * x[i];
+      }
+      y[j] = dummy;
+    }
+#else
+    for (int i=0; i<nrow; i++) y[i]=0.0;
+    for (int k=0, i=0; i<ncol; i++) { 
+      for (int j=0; j<nrow; j++) {
 	y[j] += A[k++] * x[i];
       }
     }
+#endif  
   }
 }
 
 
 void Ax(double *A, double*x1, double*x2, int nrow, int ncol, double *y1,
 	double *y2) {
-  int i,j,k;
   if (A == NULL) {
     if (nrow != ncol || nrow <= 0) BUG;
     MEMCOPY(y1, x1, sizeof(double) * nrow);
     MEMCOPY(y2, x2, sizeof(double) * nrow);
   } else {
-    for (i=0; i<nrow; i++) y1[i]=y2[i]=0.0;
-    for (k=i=0; i<ncol; i++) { 
-      for (j=0; j<nrow; j++) {
+    for (int i=0; i<nrow; i++) y1[i]=y2[i]=0.0;
+    for (int k=0, i=0; i<ncol; i++) { 
+      for (int j=0; j<nrow; j++) {
 	y1[j] += A[k] * x1[i];
 	y2[j] += A[k++] * x2[i];
       }
@@ -140,38 +146,43 @@ void Ax(double *A, double*x1, double*x2, int nrow, int ncol, double *y1,
 
 
 double XkCXtl(double *X, double *C, int nrow, int dim, int k, int l) {
-  // (k-th row of X) * C * (l-th column of X)
-    double *pX, *pY, scalar, result;
-    int i,j,ci,
-	size = nrow * dim;
-   
-  pX = X + k;
-  pY = X + l;
-  result = 0.0;
-  for (ci=0, j=0; j<size; j+=nrow) {
-      for (i=0, scalar = 0.0; i<size; i+=nrow) {
-	 scalar += pX[i] * C[ci++];
-     }
-     result += scalar * pY[j];
-  }  
+  // (k-th row of X) * C * (l-th row of X)
+  // X is nrow x dim matrix
+  // C is dim x dim matrix
+  double
+    *pX = X + k, 
+    *pY = X + l, 
+    result = 0.0;
+  int size = nrow * dim;
+  
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) reduction(+:result)
+#endif
+  for (int j=0; j<size; j+=nrow) {
+    double scalar = 0.0;
+    int ci = j * dim;
+    for (int i=0; i<size; i+=nrow) scalar += pX[i] * C[ci++];
+    result += scalar * pY[j];
+  }
   return result;
 }
 
 
 void XCXt(double *X, double *C, double *V, int nrow, int dim /* dim of C */) {
-    double *pX, *endpX, *dummy, *pdummy, scalar;
-    int i, cd, rv, ci, cv,
-      size = nrow * dim;
-
-  dummy = (double*) MALLOC(sizeof(double) * nrow * dim);
-  if (dummy == NULL) ERR("XCXt: memory allocation error in XCXt");
-  
-
-  // dummy = XC
-  for (pX = X, pdummy=dummy, endpX = pX + nrow;
-       pX < endpX; pX++, pdummy++) {
-    for (ci=0, cd=0; cd<size; cd+=nrow) {
-      for (i=0, scalar = 0.0; i<size; i+=nrow) {
+  int size = nrow * dim;
+  double  
+    *endpX = X + nrow,
+    *dummy = (double*) MALLOC(sizeof(double) * size); // dummy = XC
+  if (dummy == NULL) RFERROR("XCXt: memory allocation error in XCXt");
+ 
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+  for (double *pX = X; pX < endpX; pX++) {
+    double *pdummy = dummy + (pX - X);
+    for (int ci=0, cd=0; cd<size; cd+=nrow) {
+      double scalar = 0.0;
+      for (int i=0; i<size; i+=nrow) {
         scalar += pX[i] * C[ci++];
       }
       pdummy[cd] = scalar;
@@ -179,14 +190,18 @@ void XCXt(double *X, double *C, double *V, int nrow, int dim /* dim of C */) {
   }
 
   // V = dummy X^t
-  for (rv=0; rv<nrow; rv++) {
-    for (cv=rv; cv<nrow; cv++) {
-      for (scalar=0.0, i=0; i<size; i+=nrow) {
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+  for (int rv=0; rv<nrow; rv++) {
+    for (int cv=rv; cv<nrow; cv++) {
+      double scalar=0.0;
+      for (int i=0; i<size; i+=nrow) {
 	scalar += dummy[rv + i] * X[cv + i];
      }
       V[rv + cv * nrow] = V[cv + rv * nrow] = scalar;
     }
-  } 
+  }
 
   UNCONDFREE(dummy);
 }
@@ -194,37 +209,56 @@ void XCXt(double *X, double *C, double *V, int nrow, int dim /* dim of C */) {
 
 double xUy(double *x, double *U, double *y, int dim) {
   // U a symmetric matrix given by its upper triangular part
-  double dummy,
-    xVy = 0.0;
-  int j, d, i,
-    dimM1 = dim - 1;
-  for (j=d=0; d<dim; d++) {
-    j = dim * d;
-    for (dummy = 0.0, i=0; i<=d; i++) {
-      dummy += x[i] * U[j++];
-    }
-    for (j += dimM1; i<dim; i++, j+=dim) {
-      dummy += x[i] * U[j];
-    }
+  double xVy = 0.0;
+  int    dimM1 = dim - 1;
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) if (MULTIMINSIZE(dim)) reduction(+:xVy) 
+#endif  
+  for (int d=0; d<dim; d++) {
+    int i, 
+      j = dim * d;
+    double dummy = 0.0;
+    for (i=0; i<=d; i++) dummy += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
     xVy += dummy * y[d];
   }
   return xVy;
 }
 
-double xUxz(double *x, double *U, int dim, double *z) {
+/*
+
   // U a symmetric matrix given by its upper triangular part
-  double dummy,
-    xVx = 0.0;
-  int j, d, i,
-    dimM1 = dim - 1;
-  for (j=d=0; d<dim; d++) {
-    j = dim * d;
-    for (dummy = 0.0, i=0; i<=d; i++) {
-      dummy += x[i] * U[j++];
-    }
-    for (j += dimM1; i<dim; i++, j+=dim) { 
-      dummy += x[i] * U[j];
-    }
+  assert(z != NULL);
+  int   dimM1 = dim - 1;
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) if (MULTIMINSIZE(dim))
+#endif  
+  for (int d=0; d<dim; d++) {
+    double dummy;
+    int i,
+      j = dim * d;
+    for (dummy = 0.0, i=0; i<=d; i++) dummy += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
+    if (z!=NULL) z[d] = dummy;
+  }
+  double xVx;
+  SCALAR_PROD(z, x, dim, xVx);
+  return xVx;
+
+ */
+
+double xUxz(double *x, double *U, int dim, double *z) {
+ double xVx = 0.0;
+  int dimM1 = dim - 1;
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) reduction(+:xVx)
+#endif
+  for (int d=0; d<dim; d++) {
+    int i, 
+      j = dim * d;
+    double dummy = 0.0;
+    for (dummy = 0.0, i=0; i<=d; i++) dummy += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
     if (z != NULL) z[d] = dummy;
     xVx += dummy * x[d];
   }
@@ -237,18 +271,17 @@ double xUx(double *x, double *U, int dim) {
 
 double x_UxPz(double *x, double *U, double *z, int dim) {
 // x^t (Ux + z); U dreieckmatrix
-  double dummy,
-    xVx = 0.0;
-  int j, d, i,
-    dimM1 = dim - 1;
-  for (j=d=0; d<dim; d++) {
-    j = dim * d;
-    for (dummy = z[d], i=0; i<=d; i++) {
-      dummy += x[i] * U[j++];
-    }
-    for (j += dimM1; i<dim; i++, j+=dim) {
-      dummy += x[i] * U[j];
-    }
+  double xVx = 0.0;
+  int    dimM1 = dim - 1;
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) reduction(+:xVx)
+#endif
+  for (int d=0; d<dim; d++) {
+    int i,
+      j = dim * d;
+    double dummy = z[d];
+    for (i=0; i<=d; i++) dummy += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
     xVx += dummy * x[d];
   }
   return xVx;
@@ -256,86 +289,107 @@ double x_UxPz(double *x, double *U, double *z, int dim) {
 
 
 
-void matmult(double *A, double *B, double *C, int l, int m, int n) {
+void matmult(double *a, double *b, double *c, int l, int m, int n) {
 // multiplying an lxm- and an mxn-matrix, saving result in C
-   int i, j, k;
-   for(i=0; i<l; i++, A++, C++) {
-     for(j=0; j<n; j++) {
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+   for (int i=0; i<l; i++) {
+     double *A = a + i,
+       *C = c + i;
+     for (int j=0; j<n; j++) {
        double dummy = 0.0,
-	 *Bjm = B + j * m;
-       for(k=0; k<m; k++) dummy += A[k*l]*Bjm[k];
-       C[j*l] = dummy;
+	 *B = b + j * m;
+       for (int k=0; k<m; k++) dummy += A[k*l] * B[k];
+       C[j * l] = dummy;
      }
    }
 }
 
-void matmulttransposed(double *A, double *B, double *C, int m, int l, int n) {
+
+void matmulttransposed(double *A, double *B, double *c, int m, int l, int n) {
 // multiplying t(A) and B with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
-  int i, j;
-  for(i=0; i<l; i++, C++) {
-    double *Aim = A + i * m;
-    for(j=0; j<n; j++) C[j * l] = scalar(Aim, B + j * m, m);
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+  for (int i=0; i<l; i++) {    
+    double *C = c + i,
+      *Aim = A + i * m;
+    for (int j=0; j<n; j++) C[j * l] = SCALAR(Aim, B + j * m, m);
   }
 }
 
 
 
-void matmult_2ndtransp(double *A, double *B, double *C, int m, int l, int n) {
-// multiplying A and t(B) with dim(A)=(m,l) and dim(B)=(n, l),
+void matmult_2ndtransp(double *a, double *B, double *c, int l, int m, int n) {
+// multiplying A and t(B) with dim(A)=(l, m) and dim(B)=(n, m),
 // saving result in C
-  int i, j, k,
-    msq = m  * m;
-  for(i=0; i<l; i++, C++, A++)
-     for(j=0; j<n; j++) {
+  int msq = m  * m;
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) if (l * m * n > 1000)
+#endif
+  for (int i=0; i<l; i++) {
+    double *C = c + i,
+      *A = a + i;
+    for (int j=0; j<n; j++) {
        double dummy = 0.0,
 	 *Bj = B + j;
-       for(k=0; k<msq; k+=m) dummy += A[k] * Bj[k];
+       for (int k=0; k<msq; k+=m) dummy += A[k] * Bj[k];
        C[j*l] = dummy;
-     }
+    }
+  }
 }
 
 
 
-void matmult_tt(double *A, double *B, double *C, int m, int l, int n) {
+void matmult_tt(double *a, double *B, double *c, int m, int l, int n) {
 // calculating t(A B) with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
-   int i, j, k;
-   for(i=0; i<l; i++, A++, C += l)
-     for(j=0; j<n; j++) {
-       double dummy = 0.0,
-	 *Bjm = B + j * m;
-	for(k=0; k<m; k++) dummy += A[k * l] * Bjm[k];
-	C[j] = dummy;
-     }
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+  for (int i=0; i<l; i++) {
+    double *A = a + i,
+      *C = c + i * l;
+    for (int j=0; j<n; j++) {
+      double dummy = 0.0,
+	*Bjm = B + j * m;
+      for (int k=0; k<m; k++) dummy += A[k * l] * Bjm[k];
+      C[j] = dummy;
+    }
+  }
 }
 
 
 
 void Xmatmult(double *A, double *B, double *C, int l, int m, int n) {
 // multiplying an lxm- and an mxn-matrix, saving result in C
-  int i, j, k, jl, jm, kl, endfor;
-  double dummy;
-  for(i=0; i<l; i++) {
-    for(jl=i, jm=j=0; j<n; j++, jl+=l, jm+=m) {
-       dummy = 0.0;
-       endfor = jm + m;
-       for(kl=i, k=jm; k<endfor; k++, kl+=l) dummy += A[kl] * B[k]; 
-       C[jl] = dummy;
-     }
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+  for (int i=0; i<l; i++) {
+    for (int jl=i, jm=0, j=0; j<n; j++, jl+=l, jm+=m) {
+      double dummy = 0.0;
+      int endfor = jm + m;
+      for (int kl=i, k=jm; k<endfor; k++, kl+=l) dummy += A[kl] * B[k]; 
+      C[jl] = dummy;
+    }
   }
 }
 
 void Xmatmulttransposed(double *A, double *B, double *C, int m, int l, int n) {
 // multiplying t(A) and B with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
-  int i, j, k, jl, im, jm, endfor, jmk;
-  double dummy;
-  for(im=i=0; i<l; i++, im+=m) {
-    for(jl=i, jm=j=0; j<n; j++, jl+=l, jm+=m) {
-      dummy = 0.0;
-      endfor = im + m;
-      for(jmk=jm, k=im; k<endfor; k++) dummy += A[k] * B[jmk++]; 
+#ifdef DO_PARALLEL
+#pragma omp parallel for num_threads(CORES) 
+#endif
+  for (int i=0; i<l; i++) {
+    int im = i * m;
+    for (int jl=i, jm=0, j=0; j<n; j++, jl+=l, jm+=m) {
+      double dummy = 0.0;
+      int endfor = im + m;
+      for (int jmk=jm, k=im; k<endfor; k++) dummy += A[k] * B[jmk++]; 
       C[jl] = dummy;
     }
   }
@@ -353,13 +407,13 @@ double *matrixmult(double *m1, double *m2, int dim1, int dim2, int dim3) {
 SEXP TooLarge(int *n, int l){
 #define nTooLarge 2 // mit op
   const char *tooLarge[nTooLarge] = {"size", "msg"};
-  int i;
   SEXP namevec, info;
   PROTECT(info=allocVector(VECSXP, nTooLarge));
   PROTECT(namevec = allocVector(STRSXP, nTooLarge));
-  for (i=0; i<nTooLarge; i++) SET_STRING_ELT(namevec, i, mkChar(tooLarge[i]));
+  for (int i=0; i<nTooLarge; i++)
+    SET_STRING_ELT(namevec, i, mkChar(tooLarge[i]));
   setAttrib(info, R_NamesSymbol, namevec);
-  i=0;
+  int i=0;
   SET_VECTOR_ELT(info, i++, Int(n, l, l));
   SET_VECTOR_ELT(info, i,
 		 mkString("too many elements - increase max.elements"));
@@ -378,13 +432,12 @@ SEXP TooSmall(){
 
 
 SEXP Int(int *V, int n, int max) {
-  int i;
   SEXP dummy;
   if (V==NULL) return allocVector(INTSXP, 0);
   if (n>max) return TooLarge(&n, 1);
    if (n<0) return TooSmall();
    PROTECT(dummy=allocVector(INTSXP, n));
-  for (i=0; i<n; i++) INTEGER(dummy)[i] = V[i];
+  for (int i=0; i<n; i++) INTEGER(dummy)[i] = V[i];
   UNPROTECT(1);
   return dummy;
 }
@@ -395,13 +448,12 @@ SEXP Int(int* V, int n) {
 
 
 SEXP Logic(bool* V, int n, int max) {
-  int i;
   SEXP dummy;
   if (V==NULL) return allocVector(VECSXP, 0);
   if (n>max) return TooLarge(&n, 1);
   if (n<0) return TooSmall();
   PROTECT(dummy=allocVector(LGLSXP, n));
-  for (i=0; i<n; i++) LOGICAL(dummy)[i] = V[i];
+  for (int i=0; i<n; i++) LOGICAL(dummy)[i] = V[i];
   UNPROTECT(1);
   return dummy;
 }
@@ -410,13 +462,12 @@ SEXP Logic(bool* V, int n) {
 }
 
 SEXP Num(double* V, int n, int max) {
-  int i;
   SEXP dummy;
   if (V==NULL) return allocVector(REALSXP, 0);
   if (n>max) return TooLarge(&n, 1);
    if (n<0) return TooSmall();
   PROTECT(dummy=allocVector(REALSXP, n));
-  for (i=0; i<n; i++) REAL(dummy)[i] = V[i];
+  for (int i=0; i<n; i++) REAL(dummy)[i] = V[i];
   UNPROTECT(1);
   return dummy;
 }
@@ -425,13 +476,12 @@ SEXP Num(double* V, int n) {
 }
 
 SEXP Result(double* V, int n, int max) {
-  int i;
   SEXP dummy;
   if (V==NULL) return allocVector(REALSXP, 0);
   if (n>max) return TooLarge(&n, 1);
    if (n<0) return TooSmall();
  PROTECT(dummy=allocVector(REALSXP, n));
-  for (i=0; i<n; i++) REAL(dummy)[i] = (double) V[i];
+  for (int i=0; i<n; i++) REAL(dummy)[i] = (double) V[i];
   UNPROTECT(1);
   return dummy;
 }
@@ -441,13 +491,12 @@ SEXP Result(double* V, int n) {
 }
 
 SEXP Char(const char **V, int n, int max) {
-  int i;
   SEXP dummy;
   if (V==NULL) return allocVector(STRSXP, 0);
   if (n>max) return TooLarge(&n, 1);
    if (n<0) return TooSmall();
    PROTECT(dummy=allocVector(STRSXP, n));
-   for (i=0; i<n; i++){
+   for (int i=0; i<n; i++){
      SET_STRING_ELT(dummy, i, mkChar(V[i]));  
    }
   UNPROTECT(1);
@@ -459,9 +508,8 @@ SEXP Char(const char **V, int n) {
 }
 
 SEXP Mat(double* V, int row, int col, int max) {
-  int i, n;
   if (V==NULL) return allocMatrix(REALSXP, 0, 0);
-  n = row * col;
+  int n = row * col;
   if (n>max) {
     int nn[2];
     nn[0] = row;
@@ -470,7 +518,7 @@ SEXP Mat(double* V, int row, int col, int max) {
   }
   SEXP dummy;
   PROTECT(dummy=allocMatrix(REALSXP, row, col));
-  for (i=0; i<n; i++) REAL(dummy)[i] = V[i];
+  for (int i=0; i<n; i++) REAL(dummy)[i] = V[i];
   UNPROTECT(1);
   return dummy;
 }
@@ -481,9 +529,8 @@ SEXP Mat(double* V, int row, int col) {
 
 
 SEXP Mat_t(double* V, int row, int col, int max) {
-  int i,j, k,n;
   if (V==NULL) return allocMatrix(REALSXP, 0, 0);
-  n = row * col;
+  int n = row * col;
   if (n>max) {
     int nn[2];
     nn[0] = row;
@@ -492,8 +539,8 @@ SEXP Mat_t(double* V, int row, int col, int max) {
   }
   SEXP dummy;
   PROTECT(dummy=allocMatrix(REALSXP, row, col));
-  for (k=j=0; j<col; j++) {
-     for (i=0; i<row; i++) {
+  for (int k=0, j=0; j<col; j++) {
+     for (int i=0; i<row; i++) {
       REAL(dummy)[k++] = V[j + col * i];
     }
   }
@@ -506,10 +553,30 @@ SEXP Mat_t(double* V, int row, int col) {
 }
 
 
+SEXP MatString(char **V, int row, int col, int max) {
+  if (V==NULL) return allocMatrix(STRSXP, 0, 0);
+  int n = row * col;
+  if (n>max) {
+    int nn[2];
+    nn[0] = row;
+    nn[1] = col;
+    return TooLarge(nn, 2);
+  }
+  SEXP dummy;
+  PROTECT(dummy=allocMatrix(STRSXP, row, col));
+  for (int k=0; k<n; k++)
+    SET_STRING_ELT(dummy, k, mkChar(V[k]));
+  UNPROTECT(1);
+  return dummy;
+}
+
+SEXP MatString(char** V, int row, int col) {
+  return MatString(V, row, col, MAXINT);
+}
+
 SEXP MatInt(int* V, int row, int col, int max) {
-  int i, n;
   if (V==NULL) return allocMatrix(INTSXP, 0, 0);
-  n = row * col;
+  int n = row * col;
   if (n>max) {
     int nn[2];
     nn[0] = row;
@@ -518,7 +585,7 @@ SEXP MatInt(int* V, int row, int col, int max) {
   }
   SEXP dummy;
   PROTECT(dummy=allocMatrix(INTSXP, row, col));
-  for (i=0; i<n; i++) INTEGER(dummy)[i] = V[i];
+  for (int i=0; i<n; i++) INTEGER(dummy)[i] = V[i];
   UNPROTECT(1);
   return dummy;
 }
@@ -528,10 +595,10 @@ SEXP MatInt(int* V, int row, int col) {
 }
 
 SEXP Array3D(double** V, int depth, int row, int col, int max) {
-  int i, j, m, n;
   if (V==NULL) return alloc3DArray(REALSXP, 0, 0, 0);
-  m = row * col;
-  n = row * col * depth;
+  int
+    m = row * col,
+    n = row * col * depth;
   if (n>max) {
     int nn[3];
     nn[0] = row;
@@ -541,8 +608,8 @@ SEXP Array3D(double** V, int depth, int row, int col, int max) {
   }
   SEXP dummy;
   PROTECT(dummy=alloc3DArray(REALSXP, depth, row, col));
-  for(j=0; j<depth; j++) {
-    for (i=0; i<m; i++) {
+  for (int j=0; j<depth; j++) {
+    for (int i=0; i<m; i++) {
       REAL(dummy)[j*m+i] = V[j][i];
     }
   }
@@ -557,13 +624,19 @@ SEXP Array3D(double** V, int depth, int row, int col) {
 
 
 
-//static int ZZ = 0;
+usr_bool UsrBoolRelaxed(SEXP p, char *name, int idx) {
+  double dummy = Real(p, name, idx);
+  if (!R_finite(dummy)) return Nan;
+  return dummy==0.0 ? False : True ;
+}
+
+
 usr_bool UsrBool(SEXP p, char *name, int idx) {
   double dummy = Real(p, name, idx);
   if (dummy == 0.0) return False;
   else if (dummy == 1.0) return True;
-  else if (ISNA(dummy) || ISNAN(dummy)) return Nan;
-  ERR("invalid value for boolean variable");
+  else if (ISNAN(dummy)) return Nan;
+  RFERROR2("invalid value (%d) for boolean variable '%.50s'.", (int) dummy, name);
 }
 
 
@@ -578,13 +651,12 @@ SEXP String(char *V) {
 }
 
 SEXP String(char V[][MAXCHAR], int n, int max) {
-  int i;
   SEXP str;
-  if (V==NULL) return allocVector(VECSXP, 0);
+  if (V==NULL) return allocVector(STRSXP, 0);
   if (n>max) return TooLarge(&n, 1);
   if (n<0) return TooSmall();
   PROTECT(str = allocVector(STRSXP, n)); 
-  for (i=0; i<n; i++) {
+  for (int i=0; i<n; i++) {
     SET_STRING_ELT(str, i, mkChar(V[i]));
   }
   UNPROTECT(1);
@@ -593,23 +665,25 @@ SEXP String(char V[][MAXCHAR], int n, int max) {
 
 
 SEXP String(int *V, const char * List[], int n, int endvalue) {
-  int i,k;
   SEXP str;
-  if (V==NULL || n <= 0) return allocVector(VECSXP, 0);
+  if (V==NULL || n <= 0) return allocVector(STRSXP, 0);
+  int k;
   for (k=0; k<n; k++) {
     if (V[k] == endvalue) break;
   }
+  //  printf("k=%d, n=%d\n", k, n);
   PROTECT(str = allocVector(STRSXP, k)); 
-  for (i=0; i<k; i++) {
+  for (int i=0; i<k; i++) {
+    //  printf("V[%d]=%d\n", i, V[i]);
+    //    printf("%.50s\n", List[V[i]]);
     SET_STRING_ELT(str, i, mkChar(List[V[i]]));
   }
   UNPROTECT(1);
   return str;
 }
 
-//static int ZZ = 0;
 double Real(SEXP p, char *name, int idx) {
-  // if(++ZZ==65724){printf("type=%d %d '%s'\n",ZZ,TYPEOF(p), CHAR(STRING_ELT(p,0)));cov_model *cov;crash(cov);}
+  //  {printf("%.50s type=%d \n", name,TYPEOF(p));}
   if (p != R_NilValue) {
     assert(idx < length(p));
     switch (TYPEOF(p)) {
@@ -621,19 +695,19 @@ double Real(SEXP p, char *name, int idx) {
     default : {}
     }
   }
-  ERR2("'%s' cannot be transformed to double! (type=%d)\n", name, TYPEOF(p));  
+
+ RFERROR2("'%.50s' can not be transformed to double! (type=%d)\n", name, TYPEOF(p));  
   return RF_NA;  // to avoid warning from compiler
 }
 
 
 
 void Real(SEXP el,  char *name, double *vec, int maxn) {
-   int i, j, n;
-   if (el == R_NilValue) {
-     ERR1("'%s' cannot be transformed to double.\n", name);
+  if (el == R_NilValue) {
+    RFERROR1("'%.50s' cannot be transformed to double.\n", name);
   }
-  n = length(el);
-  for (j=i=0; i<maxn; i++) {
+  int n = length(el);
+  for (int j=0, i=0; i<maxn; i++) {
     vec[i] = Real(el, name, j);
     if (++j >= n) j=0;
   }
@@ -653,16 +727,19 @@ int Integer(SEXP p, char *name, int idx, bool nulltoNA) {
       if (ISNAN(value)) {
 	return NA_INTEGER;
       }
-      if (value == trunc(value)) return (int) value; 
+      int intvalue;
+      intvalue = (int) value;
+      //      print("%10g %d %10e\n ", value, intvalue, value-intvalue);
+      if (value == intvalue) return intvalue; 
       else {
-	ERR2("%s: integer value expected. Got %e.", name, value);
+	RFERROR2("%.50s: integer value expected. Got %10e.", name, value);
       }
     case LGLSXP :
       return  LOGICAL(p)[idx]==NA_LOGICAL ? NA_INTEGER : (int) LOGICAL(p)[idx];
     default : {}
     }
   } else if (nulltoNA) return NA_INTEGER;
-  ERR2("%s: unmatched type of parameter [type=%d]", name, TYPEOF(p));
+  RFERROR2("%.50s: unmatched type of parameter [type=%d]", name, TYPEOF(p));
   return NA_INTEGER; // compiler warning vermeiden
 }
 
@@ -672,12 +749,11 @@ int Integer(SEXP p, char *name, int idx) {
 
 
 void Integer(SEXP el, char *name, int *vec, int maxn) {
-  int i, j, n;
   if (el == R_NilValue) {
-    ERR1("'%s' cannot be transformed to integer.\n",name);
+    RFERROR1("'%.50s' cannot be transformed to integer.\n",name);
   }
-  n = length(el);
-  for (j=i=0; i<maxn; i++) {
+  int n = length(el);
+  for (int j=0, i=0; i<maxn; i++) {
     vec[i] = Integer(el, name, j);
     if (++j >= n) j=0;
   }
@@ -689,18 +765,23 @@ void Integer(SEXP el, char *name, int *vec, int maxn) {
 void Integer2(SEXP el, char *name, int *vec) {
   int n;
   if (el == R_NilValue || (n = length(el))==0) {
-      ERR1("'%s' cannot be transformed to integer.\n",name);
+      RFERROR1("'%.50s' cannot be transformed to integer.\n",name);
   }
  
   vec[0] = Integer(el, name, 0);
+  if (vec[0] == NA_INTEGER || vec[0] < 1)
+    RFERROR1("first component of '%.50s' must be at least 1", name);
   if (n==1) vec[1] = vec[0];
   else {
-    vec[1] = Integer(el, name, n-1);
+    vec[1] = Integer(el, name, n-1);    
+    if ( vec[1] != NA_INTEGER && vec[1] < vec[0])
+      RFERROR1("'%.50s' must be increasing", name);
     if (n > 2) {
-      int i, 
-	v = vec[0] + 1;
-      for (i = 1; i<n; i++, v++)
-	if (Integer(el, name, i) != v) ERR("not a sequence of numbers"); 
+      int v = vec[0] + 1;
+      for (int i = 1; i<n; i++, v++)
+	if (Integer(el, name, i) != v)
+	  RFERROR1("'%.50s' is not a sequence of numbers",name); 
+
     }
   }
 }
@@ -719,7 +800,7 @@ bool Logical(SEXP p, char *name, int idx) {
     case LGLSXP : return LOGICAL(p)[idx];
     default : {}
     }
-  ERR1("'%s' cannot be transformed to logical.\n", name);  
+  RFERROR1("'%.50s' cannot be transformed to logical.\n", name);  
   return NA_LOGICAL;  // to avoid warning from compiler
 }
 
@@ -731,37 +812,35 @@ char Char(SEXP el, char *name) {
   if (type == CHARSXP) return CHAR(el)[0];
   if (type == STRSXP) {
     if (length(el)==1) {
-      if (strlen(CHAR(STRING_ELT(el,0))) == 1)
+      if (STRLEN(CHAR(STRING_ELT(el,0))) == 1)
 	return (CHAR(STRING_ELT(el,0)))[0];
-      else if (strlen(CHAR(STRING_ELT(el,0))) == 0)
+      else if (STRLEN(CHAR(STRING_ELT(el,0))) == 0)
 	return '\0';
     }
   }
  
  ErrorHandling:
-  ERR1("'%s' cannot be transformed to character.\n",  name);  
+  RFERROR1("'%.50s' cannot be transformed to character.\n",  name);  
   return 0; // to avoid warning from compiler
 }
 
 
 void String(SEXP el, char *name, char names[][MAXCHAR], int maxlen) {
-  int i,
-    l = length(el);
-  char msg[200];
+  int l = length(el);
   SEXPTYPE type;  
   if (el == R_NilValue) goto ErrorHandling;
   if (l > maxlen)  {
-    ERR1("number of variable names exceeds %d. Take abbreviations?", maxlen);
+    RFERROR1("number of variable names exceeds %d. Take abbreviations?", maxlen);
   }
   type = TYPEOF(el);
   //  printf("type=%d %d %d %d\n", TYPEOF(el), INTSXP, REALSXP, LGLSXP);
   if (type == CHARSXP) {
-    for (i=0; i<l; i++) {
+    for (int i=0; i<l; i++) {
       names[i][0] = CHAR(el)[i];
       names[i][1] = '\0';
     }
   } else if (type == STRSXP) {
-    for (i=0; i<l; i++) {
+    for (int i=0; i<l; i++) {
       //print("%d %d\n", i, l);
       strcopyN(names[i], CHAR(STRING_ELT(el, i)), MAXCHAR);
     }
@@ -769,8 +848,7 @@ void String(SEXP el, char *name, char names[][MAXCHAR], int maxlen) {
   return;
  
  ErrorHandling:
-  sprintf(msg, "'%s' cannot be transformed to character.\n",  name);  
-  ERR(msg);
+  RFERROR1("'%.50s' cannot be transformed to character.\n",  name);  
 }
 
 
@@ -780,7 +858,7 @@ double NonNegInteger(SEXP el, char *name) {
   num = INT;
   if (num<0) {
     num=0; 
-    WARN1("'%s' which has been negative is set 0.\n",name);
+    WARN1("'%.50s', which has been negative, is set 0.\n",name);
   }
   return num;
 }
@@ -790,7 +868,7 @@ double NonNegReal(SEXP el, char *name) {
   num = NUM;
   if (num<0.0) {
     num=0.0; 
-    WARN1("%s which has been negative is set 0.\n",name);
+    WARN1("%.50s, which has been negative, is set 0.\n",name);
    }
   return num;
 }
@@ -800,7 +878,7 @@ double NonPosReal(SEXP el, char *name) {
   num = NUM;
   if (num>0.0) {
     num=0.0; 
-    WARN1("%s which has been positive is set 0.\n",name);
+    WARN1("%.50s, which has been positive, is set 0.\n",name);
   }
   return num;
 }
@@ -809,8 +887,9 @@ double PositiveInteger(SEXP el, char *name) {
   int num;
   num = INT;
   if (num<=0) {
-    num=0; 
-    WARN1("'%s' which has been negative is set 0.\n",name);
+    WARN2("'%.50s', which has been %.50s, is set 1.\n",
+	  name, num==0L ? "0" : "negative");
+    num=1L;
   }
   return num;
 }
@@ -819,8 +898,9 @@ double PositiveReal(SEXP el, char *name) {
   double num;
   num = NUM;
   if (num<=0.0) {
-    num=0.0; 
-    WARN1("%s which has been negative is set 0.\n",name);
+     WARN2("'%.50s', which has been %.50s, is set 1.\n",
+	   name, num==0.0 ? "0" : "negative");
+     num=1.0; 
    }
   return num;
 }
@@ -831,10 +911,6 @@ SEXP ExtendedInteger(double x) {
   return ScalarInteger(R_FINITE(x) ? x : NA_INTEGER);
 }
 
-SEXP ExtendedBoolean(double x) {
-  return ScalarLogical(ISNAN(x) ? NA_LOGICAL : x != 0.0);
-}
-
 SEXP ExtendedBooleanUsr(usr_bool x) {
   return ScalarLogical((int) x);
 }
@@ -846,14 +922,14 @@ int Match(char *name, name_type List, int n) {
   unsigned int ln;
   int Nr;
   Nr=0;
-  ln=strlen(name);
-  //  print("Match %d %d %s %s %d\n", Nr, n, name, List[Nr], ln);
+  ln=STRLEN(name);
+  //  print("Match %d %d %.50s %.50s %d\n", Nr, n, name, List[Nr], ln);
 
-  while ( Nr < n  && strncmp(name, List[Nr], ln)) {
+  while ( Nr < n  && STRNCMP(name, List[Nr], ln)) {
     Nr++;
   }
   if (Nr < n) { 
-    if (ln==strlen(List[Nr])) // exactmatching -- take first -- changed 1/7/07
+    if (ln==STRLEN(List[Nr])) // exactmatching -- take first -- changed 1/7/07
       return Nr;
     // a matching function is found. Are there other functions that match?
     int j; 
@@ -861,9 +937,9 @@ int Match(char *name, name_type List, int n) {
     j=Nr+1; // if two or more covariance functions have the same name 
     //            the last one is taken 
     while (j<n) {
-      while ( (j<n) && strncmp(name, List[j], ln)) {j++;}
+      while ( (j<n) && STRNCMP(name, List[j], ln)) {j++;}
       if (j<n) {
-	if (ln==strlen(List[j])) { // exactmatching -- take first 
+	if (ln==STRLEN(List[j])) { // exactmatching -- take first 
 	  return j;
 	}
 	else {multiplematching=true;}
@@ -882,17 +958,17 @@ int Match(char *name, const char * List[], int n) {
   unsigned int ln;
   int Nr;
   Nr=0;
-  ln=strlen(name);
-  //    print("Matchx %d %d %s %s %d\n", Nr, n, name, List[Nr], ln);
+  ln=STRLEN(name);
+  //    print("Matchx %d %d %.50s %.50s %d\n", Nr, n, name, List[Nr], ln);
 
-  while ( Nr < n  && strncmp(name, List[Nr], ln)) {
-    //     print("       %d %d %s %s %d\n", Nr, n, name, List[Nr], ln);
-    //   printf("%s\n", List[Nr]);
+  while ( Nr < n  && STRNCMP(name, List[Nr], ln)) {
+    //     print("       %d %d %.50s %.50s %d\n", Nr, n, name, List[Nr], ln);
+    //   printf("%.50s\n", List[Nr]);
     Nr++;
   }
   if (Nr < n) { 
-    if (ln==strlen(List[Nr])) {// exactmatching -- take first -- changed 1/7/07
-      //      print(" found  X    %d %d %s %s %d\n", Nr, n, name, List[Nr], ln);
+    if (ln==STRLEN(List[Nr])) {// exactmatching -- take first -- changed 1/7/07
+      //      print(" found  X    %d %d %.50s %.50s %d\n", Nr, n, name, List[Nr], ln);
       return Nr;
     }
     // a matching function is found. Are there other functions that match?
@@ -901,9 +977,9 @@ int Match(char *name, const char * List[], int n) {
     j=Nr+1; // if two or more covariance functions have the same name 
     //            the last one is taken 
     while (j<n) {
-      while ( (j<n) && strncmp(name, List[j], ln)) {j++;}
+      while ( (j<n) && STRNCMP(name, List[j], ln)) {j++;}
       if (j<n) {
-	if (ln==strlen(List[j])) { // exactmatching -- take first 
+	if (ln==STRLEN(List[j])) { // exactmatching -- take first 
 	  return j;
 	}
 	else {multiplematching=true;}
@@ -913,7 +989,7 @@ int Match(char *name, const char * List[], int n) {
     if (multiplematching) {return MULTIPLEMATCHING;}
   } else return NOMATCHING;
 
-  //    print(" found      %d %d %s %s %d\n", Nr, n, name, List[Nr], ln);
+  //    print(" found      %d %d %.50s %.50s %d\n", Nr, n, name, List[Nr], ln);
  
   return Nr;
 }
@@ -923,20 +999,20 @@ int Match(char *name, const char * List[], int n) {
 void GetName(SEXP el, char *name, const char * List[], int n,
 	     int defaultvalue, int endvalue, int *ans, int maxlen_ans) {
   char dummy[1000];
-  int i,
+  int 
     k = 0,
     len_el = length(el);
 
   if (TYPEOF(el) == NILSXP) goto ErrorHandling;
   if (len_el > maxlen_ans) 
-    ERR2("option '%s' is too long. Maximum length is %d.", name, maxlen_ans);
+    RFERROR2("option '%.50s' is too long. Maximum length is %d.", name, maxlen_ans);
 
   if (TYPEOF(el) == STRSXP) {    
     for (k=0; k<len_el; k++) {
       ans[k] = Match((char*) CHAR(STRING_ELT(el, k)), List, n);
       if (ans[k] < 0) {
-	if (strcmp((char*) CHAR(STRING_ELT(el, k)), " ") == 0 ||
-	    strcmp((char*) CHAR(STRING_ELT(el, k)), "") == 0) {
+	if (STRCMP((char*) CHAR(STRING_ELT(el, k)), " ") == 0 ||
+	    STRCMP((char*) CHAR(STRING_ELT(el, k)), "") == 0) {
 	  goto ErrorHandling;
 	}
 	goto ErrorHandling0;
@@ -947,14 +1023,15 @@ void GetName(SEXP el, char *name, const char * List[], int n,
   }
 
 ErrorHandling0:
-  sprintf(dummy, "'%s': unknown value '%s'. Possible values are:", 
+  SPRINTF(dummy, "'%.50s': unknown value '%.50s'. Possible values are:", 
 	  name, CHAR(STRING_ELT(el, k)));
+  int i;
   for (i=0; i<n-1; i++) {
-    char msg[1000];
-    sprintf(msg, "%s '%s',", dummy, List[i]);    
-    strcpy(dummy, msg);
+    char info[1000];
+    SPRINTF(info, "%.50s '%.50s',", dummy, List[i]);    
+    STRCPY(dummy, info);
   }
-  ERR2("%s and '%s'.", dummy, List[i]);  
+  RFERROR2("%.50s and '%.50s'.", dummy, List[i]);  
  
  ErrorHandling:
   if (defaultvalue >= 0) {
@@ -963,7 +1040,7 @@ ErrorHandling0:
     return;
   }
   
-  ERR1("'%s': no value given.", name);
+  RFERROR1("'%.50s': no value given.", name);
 }
 
 int GetName(SEXP el, char *name, const char * List[], int n,
@@ -979,3 +1056,101 @@ int GetName(SEXP el, char *name, const char * List[], int n) {
 }
 
 
+double ownround(double x) { return TRUNC((x + SIGN(x) * 0.5)); }
+
+
+double lonmod(double x, double modulus) {  
+  double 
+    halfmodulus = 0.5 * modulus,
+    y = x + modulus + halfmodulus;
+  return Mod(y, modulus) - halfmodulus;
+}
+
+/*
+
+
+
+
+double intpow(double x, int p) {
+  //int p0 = p;
+  // double x0 = x;
+
+  double res = 1.0;
+  if (p < 0) {
+    p = -p;
+    x = 1.0 / x;
+  } 
+  while (p != 0) {
+    //    printf("  ... %10e %d : %10e\n" , x, p, res);
+  if (p % 2 == 1) res *= x;
+    x *= x;
+    p /= 2;
+  }
+  return res;
+}
+
+
+
+void distInt(int *X, int*N, int *Genes, double *dist) {
+    int i,j, k, di, diff, *x, *y, ve, ho, endfor,
+	n = *N,
+	nP1 = n + 1,
+	genes = *Genes;
+ 
+  x = y = X;
+  for (j=0, i=0;  j<n;  i += nP1, j++, y += genes) {
+    dist[i] = 0.0;
+    endfor = i + (n - j);
+    for (ve = i + 1, ho = i + n, x = y + genes; 
+         ve < endfor; 
+	 ve++, ho += n) {
+      for (di=0.0, k=0; k<genes; k++, x++) {
+	diff = *x - y[k];
+	di += diff * diff;
+      }
+      dist[ve] = dist[ho] = SQRT((double) di);
+    }
+  }
+}
+
+
+void vectordist(double *v, int *Dim, double *Dist, int *diag){
+  int d, dim, dr;
+  double *v1, *v2, *end;
+  bool notdiag = (*diag==0);
+  dim = Dim[0];
+  end = v + Dim[1] * dim; 
+
+//  print("%d %d %10g %10g\n", dim , Dim[0], v, end);
+
+  for (dr=0, v1=v; v1<end; v1+=dim) { // loop is one to large??
+    v2 = v1;
+    if (notdiag) {
+       v2 += dim;
+    }
+    for (; v2<end; ) {
+      for (d=0; d<dim; v2++) {
+	Dist[dr++] = v1[d++] - *v2;
+      }
+    }
+  }
+} 
+
+
+int addressbits(void VARIABLE_IS_NOT_USED *addr) {
+#ifndef RANDOMFIELDS_DEBUGGING  
+  return 0;
+#else
+  double x = (long int) addr,
+    cut = 1e9;
+  x = x - TRUNC(x / cut) * cut;
+  return (int) x;
+#endif
+
+}
+
+
+
+
+ */
+
